# Unit tests for the NOIRLab SPARCL API Client
# EXAMPLES: (do after activating venv, in sandbox/sparclclient/)
# usrpw (User Password) is intentionally left blank in these examples.
# When actually running tests, the User Password should be included.
#
#   usrpw='' python -m unittest tests.tests_api
#
#  ### Run Against DEV Server.
#  usrpw='' serverurl=http://localhost:8050 python -m unittest tests.tests_api
#  usrpw='' serverurl=http://sparcdev2.csdc.noirlab.edu:8050 python -m unittest tests.tests_api  # noqa: E501
#  showact=1 usrpw='' serverurl=http://localhost:8050 python -m unittest tests.tests_api  # noqa: E501
#
# python -m unittest  -v tests.tests_api    # VERBOSE (show what is skipped)
# python -m unittest tests.tests_api.SparclClientTest
# python -m unittest tests.tests_api.SparclClientTest.test_find_3
# python3 -m unittest tests.tests_api.AlignRecordsTest
# usrpw='' python3 -m unittest tests.tests_api.AuthTest
#
# showact=1 python -m unittest -k test_find_5 tests.tests_api
#
#  ### Run Against DEV Server.
#  usrpw='' serverurl=http://localhost:8050 python -m unittest tests.tests_api
#
#  ### Run tests Against PAT Server.
#  export serverurl=https://sparc1.datalab.noirlab.edu/
#  usrpw='' python -m unittest tests.tests_api
#
#  ### Run Against STAGE Server.
#  serverurl=https://sparclstage.datalab.noirlab.edu/
#  usrpw='' python -m unittest tests.tests_api
#
#  ### Run Against PROD Server.
#  usrpw='' serverurl=https://astrosparcl.datalab.noirlab.edu/ python -m unittest tests.tests_api  # noqa: E501

# Python library
from contextlib import contextmanager
import unittest
from unittest import skip, skipUnless, skipIf
import datetime
import requests

#!import time
from contextlib import redirect_stdout

#! from unittest mock, skipIf, skipUnless
#!import warnings
from pprint import pformat as pf
from urllib.parse import urlparse

#! from urllib.parse import urlencode

#!from unittest.mock import MagicMock, create_autospec
import os
import io

# External Packages
import numpy
import logging
import sys
import warnings

# Local Packages
from tests.utils import tic, toc
import tests.expected_pat as exp
import sparcl.exceptions as ex
import sparcl.gather_2d as sg
import sparcl.client
import sparcl.gather_2d

#! import sparcl.utils as ut


DEFAULT = "DEFAULT"
ALL = "ALL"
drs = ["BOSS-DR16"]

_DEV1 = "http://localhost:8050"  # noqa: E221
_PAT1 = "https://sparc1.datalab.noirlab.edu"  # noqa: E221
_STAGE = "https://sparclstage.datalab.noirlab.edu"  # noqa: E221
_PROD = "https://astrosparcl.datalab.noirlab.edu"  # noqa: E221

serverurl = os.environ.get("serverurl", _PROD)
showact = False
showact = showact or os.environ.get("showact") == "1"
showcurl = False
showcurl = showcurl or os.environ.get("showcurl") == "1"
clverb = False
clverb = clverb or os.environ.get("clverb") == "1"
showall = False
showall = showall or os.environ.get("showall") == "1"
if showall:
    showact = showcurl = clverb = True

usrpw = os.environ.get("usrpw")  # password for test users
show_run_context = True  # Print message about parameters of this test run


@contextmanager
def streamhandler_to_console(lggr):
    # Use 'up to date' value of sys.stdout for StreamHandler,
    # as set by test runner.
    stream_handler = logging.StreamHandler(sys.stdout)
    lggr.addHandler(stream_handler)
    yield
    lggr.removeHandler(stream_handler)


def testcase_log_console(lggr):
    def testcase_decorator(func):
        def testcase_log_console(*args, **kwargs):
            with streamhandler_to_console(lggr):
                return func(*args, **kwargs)

        return testcase_log_console

    return testcase_decorator


# Arrange to run all doctests.
# Add package paths to python files.
# The should contain testable docstrings.
def load_tests(loader, tests, ignore):
    import doctest

    if serverurl == _PROD:
        print(f"Arranging to run doctests against: sparcl.client")
        tests.addTests(doctest.DocTestSuite(sparcl.client))

        print(f"Arranging to run doctests against: sparcl.gather_2d")
        tests.addTests(doctest.DocTestSuite(sparcl.gather_2d))
    else:
        print(
            "Not running doctests since you are not running client"
            " against the PRODUCTION server."
        )

    return tests


def print_run_context(cls):
    print(
        f"""
    Running Client Tests
      against Server: \t"{urlparse(serverurl).netloc}"
      comparing to: \t{exp.__name__}
      {showact=}
      {showcurl=}
      client={cls.client}

    For REPRODUCIBLE RESULTS rebuild Server DB before running tests!
    Use: init-db.sh
    """
    )


class SparclClientTest(unittest.TestCase):
    """Test access to each endpoint of the Server API"""

    maxDiff = None  # too see full values in DIFF on assert failure
    # assert_equal.__self__.maxDiff = None

    @classmethod
    def setUpClass(cls):
        if clverb:
            print(
                f"\n# Running SparclClientTest:setUpClass() "
                f"{str(datetime.datetime.now())}!"
            )

        # Client object creation compares the version from the Server
        # against the one expected by the Client. Raise error if
        # the Client is at least one major version behind.

        cls.client = sparcl.client.SparclClient(
            url=serverurl,
            verbose=clverb,
            show_curl=showcurl,
        )
        cls.timing = dict()
        cls.doc = dict()
        cls.count = dict()

        global show_run_context

        if show_run_context:
            print_run_context(cls)
            show_run_context = False

        # Get some id_lists to use in tests
        found = cls.client.find(
            ["sparcl_id", "specid"], sort="sparcl_id", limit=5
        )
        sparc_tups, spec_tups = list(
            zip(*[(r["sparcl_id"], r["specid"]) for r in found.records])
        )
        sparc_ids, spec_ids = list(sparc_tups), list(spec_tups)
        if showact:
            print(f"sparc_ids={sparc_ids}")
            print(f"spec_ids={spec_ids}")

        # Lists of SPECID
        cls.specid_list0 = spec_ids[:2]
        cls.specid_list2 = spec_ids[:3]
        cls.specid_list5 = spec_ids[:5]
        # two real specids, one fake
        cls.specid_list3 = spec_ids[:2] + [300000000000000001]
        # two fake specids
        cls.specid_list4 = [300000000000000001, 111111111111111111]

        # Lists of SPARCL_ID (UUID)
        cls.uuid_list0 = sparc_ids[:3]
        cls.uuid_list2 = sparc_ids[:3]
        # two real UUIDs, one fake
        cls.uuid_list3 = sparc_ids[:2] + [
            "00001ebf-d030-4d59-97e5-060c47202897"
        ]
        # two (probably) fake UUIDs
        cls.uuid_list4 = [
            "00001ebf-d030-4d59-97e5-060c47202897",
            "ff1e9a12-f21a-4050-bada-a1e67a265885",
        ]
        if clverb:
            print(
                f"\n# Completed SparclClientTest:setUpClass() "
                f"{str(datetime.datetime.now())}\n"
            )

    @classmethod
    def tearDownClass(cls):
        pass

    # Full records are big.  Get the gist of them.
    def records_expected(self, recs, expd, jdata=None, show=False):
        actual = sorted(recs[0].keys())
        expected = sorted(eval(expd))  # e.g. 'ep.retrieve_1'
        if show:
            #!f'\nEXPECTED={expected}'
            print(f"{expd}: ACTUAL={actual}")
        self.assertEqual(actual, expected, "Actual to Expected")
        return actual

    #################################
    # ## Convenience Functions
    # ##

    def test_get_all_fields(self):
        """Get the intersection of all fields that are tagged as 'all'."""
        actual = self.client.get_all_fields()
        if showact:
            print(f"all_fields: actual={pf(actual)}")
        self.assertEqual(actual, exp.all_fields, msg="Actual to Expected")

    def test_get_default_fields(self):
        """Get the intersection of all fields that are tagged as 'default'."""
        actual = self.client.get_default_fields()
        if showact:
            print(f"get_default_fields: actual={pf(actual)}")
        self.assertEqual(actual, exp.default_fields, msg="Actual to Expected")

    ###
    #################################

    #################################
    # ## Convenience client Methods
    # ##
    # ##
    # ################################

    def test_missing_0(self):
        """Known missing"""
        uuids = [99, 88, 777]
        missing = self.client.missing(self.uuid_list0 + uuids)
        self.assertEqual(
            sorted(missing), sorted(uuids), msg="Actual to Expected"
        )

    def test_missing_1(self):
        """None missing"""
        # uuids = self.client.sample_uuids()
        uuids = self.uuid_list0

        missing = self.client.missing(uuids)
        if showact:
            print(f"missing_1: missing={missing}")
        self.assertEqual(missing, [], msg="Actual to Expected")

    def test_missing_specids_1(self):
        """Specids (not UUID) missing"""
        badid = "NOT_SPEC_ID"
        specids = set([badid] + self.specid_list0[:1])
        missing = set(self.client.missing_specids(specids, verbose=False))
        if showact:
            print(f"missing_specids_1: specids={specids}")
            print(f"missing_specids_1: missing={missing}")
        self.assertEqual(missing, set([badid]), msg="Actual to Expected")

    def test_retrieve_0(self):
        """Get spectra using small list of SPECIDS."""
        res = self.client.retrieve_by_specid(
            self.specid_list0,
            include=["sparcl_id", "specid", "flux"]
        )
        actual = sorted(res.records[0].keys())
        if showact:
            print(f"retrieve_0: actual={actual}")

        self.assertEqual(actual, exp.retrieve_0, msg="Actual to Expected")

    def test_retrieve_0b(self):
        """Get spectra using small list of uuids."""
        name = "retrieve_0b"
        uuids = self.uuid_list0

        tic()
        res = self.client.retrieve(uuids)
        self.timing[name] = toc()
        actual = sorted(res.records[0].keys())

        if showact:
            print(f"retrieve_0b: actual={pf(actual)}")

        self.assertEqual(actual, exp.retrieve_0b, msg="Actual to Expected")

    def test_retrieve_1(self):
        """Raise exception when unknown field referenced in include"""
        inc2 = ["bad_field_name", "flux", "or_mask"]

        #! specids = sorted(self.client.sample_specids(samples=1,
        #!                                             dataset_list=drs))
        with self.assertRaises(ex.BadInclude):
            self.client.retrieve_by_specid(
                self.specid_list0, include=inc2, dataset_list=drs
            )

    #! @skip('Cannot find an example of this edge case occuring')
    #! def test_retrieve_2(self):
    #!     """Issue warning when a Path has no value."""
    #!     inc2 = ['flux', 'spectra.coadd.OR_MASK']
    #!
    #!     specids = sorted(self.client.sample_specids(samples=1,
    #!                                           dataset_list=drs))
    #!     with self.assertWarns(Warning):
    #!         records = self.client.retrieve_by_specid(specids,
    #!                                        include=inc2,
    #!                                        dataset_list=drs)

    def test_retrieve_3(self):
        """Issue warning when some sids do not exist."""
        uuids = self.uuid_list0
        drs = ["SDSS-DR16", "BOSS-DR16", "DESI-EDR"]
        with self.assertWarns(Warning):
            self.client.retrieve(uuids + [999], dataset_list=drs)

    def test_retrieve_5(self):
        """Limit number of records returned by retrieve_by_specid."""
        res = self.client.retrieve_by_specid(
            self.specid_list5,
            include=["specid", "ivar"],
            limit=2,
        )
        actual = len(res.records)
        if showact:
            print(f"retrieve_5: actual={actual}")

        self.assertEqual(actual, exp.retrieve_5, msg="Actual to Expected")

    def test_find_0(self):
        """Get metadata using search spec."""

        outfields = ["data_release", "telescope"]
        # To get suitable constraints (in sparc-shell on Server):
        #    list(FitsRecord.objects.all().values('ra','dec'))
        constraints = {"data_release": ['BOSS-DR16']}
        found = self.client.find(outfields, constraints=constraints, limit=3)
        actual = found.records[:1]
        if showact:
            print(f"find_0: actual={pf(actual[:1])}")
        self.assertEqual(actual, exp.find_0, msg="Actual to Expected")

    @skip("fiddly bit skipped until we use factoryboy")
    def test_find_1(self):
        """Get metadata using search spec."""
        outfields = ["data_release", "specid"]
        constraints = {
            "redshift": [0.191, 0.192],
            "exptime": [2100.2, 2100.31],
            "data_release": ["SDSS-DR16"],
        }
        found = self.client.find(
            outfields, constraints=constraints, limit=1, sort="specid"
        )  # @@@
        actual = sorted(found.records, key=lambda rec: rec["specid"])
        if showact:
            print(f"find_1: actual={pf(actual)}")
        self.assertEqual(
            actual,
            sorted(exp.find_1, key=lambda rec: rec["specid"]),
            msg="Actual to Expected",
        )

    @skip("Takes too long for regression tests when used on big database.")
    def test_find_2(self):
        """Limit=None."""
        outfields = ["sparcl_id", "ra", "dec"]
        found = self.client.find(
            outfields, limit=None, sort="sparcl_id"
        )  # @@@
        actual = len(found.records)
        if showact:
            print(f"find_2: actual={pf(actual)}")
        self.assertEqual(actual, exp.find_2, msg="Actual to Expected")

    def test_find_3(self):
        """Limit=3."""
        outfields = ["data_release"]
        found = self.client.find(outfields, limit=3, sort="data_release")
        actual = sorted(found.records, key=lambda rec: rec["data_release"])
        if showact:
            print(f"find_3: actual={pf(actual)}, len(actual)={len(actual)}")
        self.assertEqual(
            len(actual),
            3,
            msg="Actual to Expected",
        )

    def test_find_4(self):
        """Check found.ids"""
        outfields = ["sparcl_id", "ra", "dec"]
        found = self.client.find(outfields, limit=3, sort="sparcl_id")  # @@@
        # Since all UUIDs have the same length, check the length of the
        # first ID to match.
        actual = len(sorted(found.ids)[0])
        if showact:
            print(f"find_4: actual={pf(actual)}")
        self.assertEqual(actual, exp.find_4, msg="Actual to Expected")

    # DLS-365
    @skip("Not implemented. Waiting for switch to ingest-time field naming ")
    def test_find_5a(self):
        """Aux field values when they exists in all found records"""
        found = self.client.find(
            ["data_release", "mjd"], limit=5, sort="sparcl_id"
        )
        actual = found.records
        if showact:
            print(f"find_5a: actual={pf(actual)}")
        self.assertEqual(actual, exp.find_5a, msg="Actual to Expected")

    # DLS-365
    @skip("Not implemented. Waiting for switch to ingest-time field naming ")
    def test_find_5b(self):
        """Aux field in one Data Set but not another. (proper subset)"""
        cons = {"data_release": ["SDSS-DR16", "DESI-EDR"]}
        f0 = self.client.find(
            ["data_release"], constraints=cons, limit=5, sort="sparcl_id"
        )
        f1 = self.client.find(
            ["data_release", "plate"],
            constraints=cons,
            limit=5,
            sort="sparcl_id",
        )
        self.assertEqual(f0.count, f1.count)

    # DLS-365
    @skip("Not implemented. Waiting for switch to ingest-time field naming ")
    def test_find_5c(self):
        """Aux field values when they do not exist in any found records"""
        cons = {"data_release": ["SDSS-DR16", "DESI-EDR"]}
        nsf = "NO_SUCH_FIELD"
        f1 = self.client.find(
            ["data_release", nsf], constraints=cons, limit=5, sort="sparcl_id"
        )
        msg = f'Expected field "{nsf}" with value None in all records'
        self.assertTrue(nsf in f1.records[0].keys(), msg)

    def test_reorder_1a(self):
        """Reorder retrieved records by sparcl_id."""
        name = "reorder_1a"
        ids = self.uuid_list2

        tic()
        res = self.client.retrieve(ids)
        self.timing[name] = toc()
        res_reorder = res.reorder(ids)
        actual = [f["sparcl_id"] for f in res_reorder.records]
        if showact:
            print(f"reorder_1a: actual={pf(actual)}")
        self.assertEqual(actual, ids, msg="Actual to Expected")

    def test_reorder_1b(self):
        """Reorder retrieved records by specid."""
        name = "reorder_1b"
        specids = self.specid_list2

        tic()
        res = self.client.retrieve_by_specid(specids)
        self.timing[name] = toc()
        res_reorder = res.reorder(specids)
        actual = [f["specid"] for f in res_reorder.records]
        if showact:
            print(f"reorder_1b: actual={pf(actual)}")
        self.assertEqual(actual, specids, msg="Actual to Expected")

    def test_reorder_2a(self):
        """Reorder records when sparcl_id is missing from database, after using
        retrieve()."""
        name = "reorder_2a"
        ids = self.uuid_list3

        tic()
        with self.assertWarns(Warning):
            res = self.client.retrieve(ids)
        self.timing[name] = toc()
        with self.assertWarns(Warning):
            res_reorder = res.reorder(ids)
        actual = [f["sparcl_id"] for f in res_reorder.records]
        if showact:
            print(f"reorder_2a: actual={pf(actual)}")
        self.assertEqual(actual, ids[:2] + ["None"], msg="Actual to Expected")

    def test_reorder_2b(self):
        """Reorder records when specid is missing from database, after
        using retrieve_by_specid()."""
        name = "reorder_2b"
        specids = self.specid_list3

        tic()
        res = self.client.retrieve_by_specid(specids)
        self.timing[name] = toc()
        with self.assertWarns(Warning):
            res_reorder = res.reorder(specids)
        actual = [f["specid"] for f in res_reorder.records]
        if showact:
            print(f"reorder_2b: actual={pf(actual)}")
        self.assertEqual(
            actual, specids[:2] + [None], msg="Actual to Expected"
        )

    def test_reorder_3a(self):
        """Test for expected Exception when a list of sparcl_ids with
        length 0 is passed to reorder method after using retrieve()."""
        name = "reorder_3a"
        ids = self.uuid_list2
        og_ids = []

        tic()
        res = self.client.retrieve(ids)
        self.timing[name] = toc()
        with self.assertRaises(ex.NoIDs):
            res.reorder(og_ids)

    def test_reorder_3b(self):
        """Test for expected Exception when a list of specids with length 0
        is passed to reorder method after using retrieve_by_specid()."""
        name = "reorder_3b"
        specids = self.specid_list2
        og_specids = []

        tic()
        res = self.client.retrieve_by_specid(specids)
        self.timing[name] = toc()
        with self.assertRaises(ex.NoIDs):
            res.reorder(og_specids)

    def test_reorder_4a(self):
        """Test for expected Exception when there are no records, using
        sparcl_ids and retrieve()."""
        name = "reorder_4a"
        ids = self.uuid_list4

        tic()
        drs = ["SDSS-DR16", "BOSS-DR16", "DESI-EDR"]
        with self.assertWarns(Warning):
            res = self.client.retrieve(ids, dataset_list=drs)
        self.timing[name] = toc()
        with self.assertRaises(ex.NoRecords):
            res.reorder(ids)

    def test_reorder_4b(self):
        """Test for expected Exception when there are no records, using specids
        and retrieve_by_specid()."""
        name = "reorder_4b"
        specids = self.specid_list4

        tic()
        drs = ["SDSS-DR16", "BOSS-DR16", "DESI-EDR"]
        res = self.client.retrieve_by_specid(specids, dataset_list=drs)
        self.timing[name] = toc()
        with self.assertRaises(ex.NoRecords):
            res.reorder(specids)

    def test_dls_468(self):
        idss = self.client.find(
            constraints={"data_release": ["SDSS-DR16"]}, limit=1
        ).ids
        re = self.client.retrieve(
            uuid_list=idss,
            include=["ra", "dec"],
            dataset_list=["SDSS-DR16"],
            verbose=False,
        )
        self.assertEqual(1, re.count)


# See DLS-280
@skip("takes to long to retrieve")  # 7/12/23
class AlignRecordsTest(unittest.TestCase):
    """Test ability to align spectra and all records by wavelength grid"""

    @classmethod
    def setUpClass(cls):
        if clverb:
            print(
                f"\n# Running AlignRecordsTest:setUpClass() "
                "{str(datetime.datetime.now())}"
            )

        cls.client = sparcl.client.SparclClient(
            url=serverurl, verbose=clverb, show_curl=showcurl
        )
        found = cls.client.find(
            constraints={"data_release": ["BOSS-DR16"]}, limit=20
        )
        cls.found = found
        cls.specflds = ["wavelength", "flux", "ivar", "mask", "model"]
        cls.got = cls.client.retrieve(found.ids, include=cls.specflds)

        if clverb:
            print(
                f"\n# Completed AlignRecordsTest:setUpClass() "
                f"{str(datetime.datetime.now())}\n"
            )

    @classmethod
    def tearDownClass(cls):
        pass

    # Requirement #1 from DLS-280
    def test_align_1(self):
        """The grid value return by align_records is a numpy array of
        Floats."""
        ar_dict, grid = sg.align_records(
            self.got.records, fields=["wavelength", "flux", "model"]
        )
        self.assertTrue(isinstance(grid, numpy.ndarray))
        self.assertTrue(isinstance(grid[0], numpy.float64))

    # Requirement #2 from DLS-280
    def test_align_2(self):
        """Allow conversion of all SPECTRA-type fields (except for wavelength)
        to be returned as same type of structure and registered to the same
        wavelength grid."""

        #! print(f'Fields={list(self.got.records[0].keys())}')
        ar_dict, grid = sg.align_records(
            self.got.records, fields=self.specflds
        )
        self.assertEqual(sorted(ar_dict.keys()), sorted(self.specflds))

    # Requirement #3 from DLS-280
    def test_align_3(self):
        """Verify shapes of arrays"""

        ar_dict, grid = sg.align_records(
            self.got.records, fields=self.specflds
        )
        shape = list(ar_dict.values())[0].shape
        self.assertEqual(shape, (20, 4621))

    # Requirement #4 from DLS-280
    # Difficult to implement
    @skip("Not implemented")
    def test_align_4(self):
        """All align fields are valid spectra fields"""
        self.assertTrue(False)

    # Requirement #5 from DLS-280
    def test_align_5(self):
        """Verify wavelength given."""
        msg = 'You must provide "wavelength" spectra field'
        with self.assertRaises(Exception, msg=msg):
            ar_dict, grid = sg.align_records(
                self.got.records, fields=["flux", "model"]
            )

    # Requirement #6  from DLS-280
    def test_align_6(self):
        """Default FIELDS to flux,wavelength."""
        got = self.client.retrieve(
            self.found.ids, include=["flux", "model", "wavelength"]
        )
        ar_dict, grid = sg.align_records(got.records)
        self.assertEqual(ar_dict["flux"].shape, (20, 4621))

    # Requirement #7  from DLS-280
    @skip("Data on DEV does not seem good enough to invoke this.")
    def test_align_7(self):
        """Error message if PRECISION not adequate for alignment"""
        # 7 precision does not support alignment
        got = self.client.retrieve(
            self.found.ids, include=["wavelength", "flux", "model"]
        )
        msg = f"bad precision"
        with self.assertRaises(Exception, msg=msg):
            ar_dict, grid = sg.align_records(got.records, precision=11)
            # shape = ar_dict['flux'].shape


@skipIf("usrpw" in os.environ, "Testing auth using usrpw env var")
class NoopTest(unittest.TestCase):
    """Non-tests."""

    def test_noop(self):
        print("<Skipped-AuthTest>", end="", flush=True)


@skipUnless("usrpw" in os.environ, "Password is required to test auth")
class AuthTest(unittest.TestCase):
    """Test authorization and authentication features"""

    @classmethod
    def setUpClass(cls):
        if clverb:
            print(
                f"\n# Running AuthTest:setUpClass() "
                f"{str(datetime.datetime.now())}"
            )

        cls.client = sparcl.client.SparclClient(
            url=serverurl, verbose=clverb, show_curl=showcurl
        )

        global show_run_context

        if show_run_context:
            print_run_context(cls)
            show_run_context = False

        cls.outflds = ["sparcl_id", "data_release"]
        cls.inc = ["data_release", "flux"]

        # Test users
        cls.auth_user = "test_user_1@noirlab.edu"
        cls.unauth_user = "test_user_2@noirlab.edu"

        # Dataset lists
        cls.Pub = ["BOSS-DR16", "DESI-DR1", "DESI-EDR", "SDSS-DR16"]
        cls.Priv = ["SDSS-DR17-test", "BOSS-DR17", "SDSS-DR17"]
        cls.PrivPub = cls.Priv + cls.Pub
        cls.PrivPub.sort()
        cls.PrivPub1 = [cls.Priv[0], cls.Pub[0]]

        # Sample list of sparcl_ids from each data set
        out = ["sparcl_id"]

        cls.cons = dict(spectype=["GALAXY"], redshift=[0.5, 0.9])

        # Silence output from login/logout
        #   (redirect only affects the WITH context)
        with redirect_stdout(io.StringIO()):  # as f:
            cls.client.login(cls.auth_user, usrpw)
        cls.uuid_priv = (
            cls.client.find(
                outfields=out,
                constraints={"data_release": [cls.Priv[0]]},
                limit=2,
                sort="sparcl_id",
            )
        ).ids
        # cls.uuid_pub = (  # cls.uuid_sdssdr16
        #    cls.client.find(
        #        outfields=out,
        #        constraints={"data_release": cls.Pub},
        #        limit=2,
        #        sort="sparcl_id",
        #    )
        # ).ids
        cls.uuid_pub = (
            cls.client.find(
                outfields=out,
                constraints={"data_release": ["BOSS-DR16"]},
                limit=2,
                sort="sparcl_id",
            )
        ).ids
        cls.uuid_privpub = cls.uuid_priv + cls.uuid_pub

        with redirect_stdout(io.StringIO()):  # as f:
            cls.client.logout()

        if clverb:
            print(  # @@@ Was not displaying
                f"\n# Completed AuthTest:setUpClass() "
                f"{str(datetime.datetime.now())}\n"
            )

    def silent_login(cls, usr, usrpw, silent=True):
        if silent:
            with redirect_stdout(io.StringIO()):  # as f:
                cls.client.login(usr, usrpw)
        else:
            cls.client.login(usr, usrpw)

    def silent_logout(cls):
        with redirect_stdout(io.StringIO()):  # as f:
            cls.client.logout()

    @classmethod
    def tearDownClass(cls):
        pass

    # curl -X 'POST' \
    #   'http://localhost:8050/sparc/get_token/' \
    #   -H 'Content-Type: application/json' \
    #   -d '{"email": "test_user_1@noirlab.edu", "password": "XX"}'; echo
    #
    # > Could not get token from SSO server:
    #   HTTPSConnectionPool(host='docker1.csdc.noirlab.edu', port=443):
    #   Max retries exceeded with url: /api/token/
    # (Caused by SSLError(SSLCertVerificationError(1,
    #                                   '[SSL: CERTIFICATE_VERIFY_FAILED]
    # certificate verify failed: unable to get local issuer certificate
    # (_ssl.c:997)')))

    def test_sso_server(self):
        sso_server = "https://sso.csdc.noirlab.edu/"
        response = requests.get(sso_server)
        self.assertEqual(response.status_code, 200, response.content.decode())

    def test_get_token(self):
        """Make sure we can get expected SSO token."""
        json = {"email": self.auth_user, "password": usrpw}
        if showact:
            print(f"test_get_token: {json=}")

        expected = 279
        res = requests.post(f"{self.client.apiurl}/get_token/", json=json)
        self.assertEqual(res.status_code, 200, res.content.decode())
        token = res.json()['access']
        actual = len(token)
        if showact:
            print(f"test_get_token: ({len(token)}) {token=!s}")
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def getToken(self):
        json = {"email": self.auth_user, "password": usrpw}
        res = requests.post(f"{self.client.apiurl}/get_token/", json=json)
        token = res.json()
        self.client.token = token['access']
        self.client.renew_token = token['refresh']

    def test_token_not_expired(self):
        self.getToken()
        expired = self.client.token_expired(renew=False)
        self.assertEqual(expired, False, "Token not expired")

    def test_token_expired(self):
        self.getToken()
        if showact:
            print(f"Token: {self.client.token=!s}")

        self.client.token_exp = datetime.datetime.now()
        expired = self.client.token_expired(renew=False)
        self.assertEqual(expired, True, "Token expired")

    def test_token_expired_renew(self):
        """
            POST http://localhost:8050/sparc/renew_token/
            Content-Type: application/json
            {
              "token": "..."
            }

        """
        self.getToken()
        self.client.token_exp = datetime.datetime.now()

        expired = self.client.token_expired(renew=True)
        self.assertEqual(expired, False, "Token expired, renewed")

    def test_authorized_1(self):
        """Test authorized method with authorized user signed in"""
        self.silent_login(self.auth_user, usrpw)
        actual = self.client.authorized
        if showact:
            print(f"1 authorized_1: {actual=}")
            #!time.sleep(1)
            #!print(f"2 authorized_1: {self.client.authorized}")
            #!time.sleep(20)
            #!print(f"3 authorized_1: {self.client.authorized}")
        self.assertEqual(actual, exp.authorized_1, msg="Actual to Expected")
        self.silent_logout()

    def test_authorized_2(self):
        """Test authorized method with unauthorized user signed in"""
        self.silent_login(self.unauth_user, usrpw)
        actual = self.client.authorized
        if showact:
            print(f"authorized_2: actual={actual}")

        self.assertEqual(actual, exp.authorized_2, msg="Actual to Expected")
        self.silent_logout()

    def test_authorized_3(self):
        """Test authorized method on anonymous user not signed in"""
        self.silent_logout()
        actual = self.client.authorized
        if showact:
            print(f"authorized_3: actual={actual}")
        self.assertEqual(actual, exp.authorized_3, msg="Actual to Expected")

    def auth_find(self, user, drs, expvar, cons={}, limit=1000):
        expected = eval(expvar)  # e.g. 'ep.retrieve_N'
        #!print(f'{expvar}: {user=} {drs=} ')
        self.silent_login(user, usrpw)
        #!print(f'{expvar}: {self.client.authorized=} {user=} {drs=} ')
        out = self.outflds
        try:
            if drs is None:
                found = self.client.find(
                    outfields=out,
                    constraints=cons,
                    limit=limit
                )
            else:
                found = self.client.find(
                    outfields=out,
                    constraints=dict(data_release=drs),
                    limit=limit,
                )
            #!if showact:
            #!    print(f"{expvar}: {found.records=}")

            #actual = sorted(set([r._dr for r in found.records]))
            actual = str(found)
        except Exception as err:
            actual = str(err)
        finally:
            self.silent_logout()
        if showact:
            print(
                f"{expvar}: {actual=}"
                f" {self.client.authorized=} {user=} {drs=}"
            )
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def auth_retrieve(self, user, drs, expvar):
        expected = eval(expvar)  # e.g. 'ep.retrieve_N'
        self.silent_login(user, usrpw)
        #!print(f'{expvar}: {self.client.authorized=} {user=} {drs=} ')
        ids = self.uuid_privpub
        inc = ["sparcl_id", "data_release"]
        try:
            if drs is None:
                got = self.client.retrieve(uuid_list=ids, include=inc)
            else:
                got = self.client.retrieve(
                    uuid_list=ids, include=inc, dataset_list=drs
                )
            actual = sorted(set([r._dr for r in got.records]))
        except Exception as err:
            actual = str(err)
        finally:
            self.silent_logout()
        if showact:
            print(
                f"{expvar}: {actual=}"
                f" {self.client.authorized=} {user=} {drs=}"
            )
        self.assertEqual(actual, expected, msg="Actual to Expected")

    # | METHOD | USER | DATASETS | OK? |
    # | find | Auth | Priv,Pub | PASS |
    # @skip("Does not return Priv")  # @@@
    def test_auth_find_1(self):
        """Test find method with authorized user; private data set specified.
        Should find Authorized Private DR"""
        exp = "exp.auth_find_1"
        self.auth_find(self.auth_user, self.PrivPub1, exp)

    # | find | Auth | None | PASS |
    # @skip("Does not return Priv")  # @@@
    def test_auth_find_2(self):
        """Test find method with authorized user; no data sets specified.
        Should find Authorized Private DR"""
        exp = "exp.auth_find_2"
        constr = {'datasetgroup': ['DESI']}
        self.auth_find(self.auth_user, None, exp, cons=constr)

    # | find | Unauth | Priv,Pub | FAIL |
    def test_auth_find_3(self):
        """Test find method with unauthorized user; private data set"""
        exp = "exp.auth_find_3"
        self.auth_find(self.unauth_user, self.PrivPub1, exp)

    # | find | Unauth | None | PASS |
    def test_auth_find_4(self):
        """Test find method with unauthorized user; no data sets specified"""
        exp = "exp.auth_find_4"
        self.auth_find(self.unauth_user, None, exp)

    # | find | Anon | Priv,Pub | FAIL |
    def test_auth_find_5(self):
        """Test find method with anonymous user; private data set
        specified"""
        exp = "exp.auth_find_5"
        self.auth_find(None, self.PrivPub1, exp)

    # | find | Anon | None | PASS |
    def test_auth_find_6(self):
        """Test find method with anonymous user; no data sets specified"""
        exp = "exp.auth_find_6"
        self.auth_find(None, None, exp)

    # | retrieve | Auth | Priv,Pub | PASS |
    # @skip("Does not return Priv")  # @@@
    def test_auth_retrieve_1(self):
        """Retrieve method with authorized user; private data set specified"""
        exp = "exp.auth_retrieve_1"
        self.auth_retrieve(self.auth_user, self.PrivPub1, exp)

    # | retrieve | Auth | None | PASS |
    # @skip("Does not return Priv")  # @@@
    def test_auth_retrieve_2(self):
        """Retrieve method with authorized user; no data sets specified.
        Should include Authorized Private DR"""
        exp = "exp.auth_retrieve_2"
        self.auth_retrieve(self.auth_user, None, exp)

    # | retrieve | Unauth | Priv,Pub | FAIL|
    def test_auth_retrieve_3(self):
        """Retrieve method with unauthorized user; private dataset specified"""
        exp = "exp.auth_retrieve_3"
        self.auth_retrieve(self.unauth_user, self.PrivPub1, exp)

    # | retrieve | Unauth | None | PASS |
    # Should PASS since default (no data sets) means only
    # retrieve from authorized datasets (Public, in this case)
    def test_auth_retrieve_4(self):
        """Retrieve method with unauthorized user; no datasets specified"""
        warnings.filterwarnings("ignore")
        exp = "exp.auth_retrieve_4"
        self.auth_retrieve(self.unauth_user, None, exp)

    # | retrieve | Unauth | Pub | PASS |
    def test_auth_retrieve_5(self):
        """Retrieve method with unauthorized user; public datasets specified"""
        warnings.filterwarnings("ignore")
        exp = "exp.auth_retrieve_5"
        self.auth_retrieve(self.unauth_user, self.Pub, exp)

    # | retrieve | Anon | Priv,Pub | FAIL |
    def test_auth_retrieve_6(self):
        """Retrieve method with anonymous user; private data set specified"""
        self.silent_logout()
        exp = "exp.auth_retrieve_6"
        self.auth_retrieve(None, self.PrivPub1, exp)
        #!# Replace exception name below once real one is created

    # | retrieve | Anon | None | PASS |
    # Should PASS since default (no data sets) means only
    # retrieve from authorized datasets (Public, in this case)
    def test_auth_retrieve_7(self):
        """Retrieve method with anonymous user; no data sets specified"""
        warnings.filterwarnings("ignore")
        exp = "exp.auth_retrieve_7"
        self.auth_retrieve(None, None, exp)

    # | retrieve | Anon | Pub | PASS |
    def test_auth_retrieve_8(self):
        """Retrieve method with anonymous user; public data sets specified"""
        warnings.filterwarnings("ignore")
        exp = "exp.auth_retrieve_8"
        self.auth_retrieve(None, self.Pub, exp)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import re
import sys
sys.path.insert(0, os.path.abspath(".."))
from sparcl import __version__

# -- Project information -----------------------------------------------------

project = "SPARCL"
copyright = "2024, S.Pothier, A.Jacques"
author = "S.Pothier, A.Jacques"

#!version = client_version
sver = __version__
mo = re.match(
    "^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$",
    sver,
)
release = sver
version = mo.group("major") + "." + mo.group("minor")
# The full version, including alpha/beta/rc tags


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_mdinclude",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "y"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["../sparcl/benchmarks/"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
autosummary_generate = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



#! /usr/bin/env python
"""Benchmark speed of SPARC spectra retrieve with various parameters.
"""
# EXAMPLES:
# cd ~/sandbox/sparclclient
# python3 -m sparcl.benchmarks.benchmarks ~/data/sparc/sids5.list
# python3 -m sparcl.benchmarks.benchmarks ~/data/sparc/sids644.list

# Alice reported 22 minutes on 64K retrieved from specClient (rate=48 spec/sec)
#   slack.spectro: 3/31/2021

# Standard Python library
import argparse
import logging
import os
from pprint import pformat

# External packages
import psutil

# Local packages
from ..client import SparclClient
from ..utils import tic, toc, here_now

# rooturl = 'http://localhost:8030/' #@@@
rooturl = "http://sparc1.datalab.noirlab.edu:8000/"


def human_size(num, units=["b", "KB", "MB", "GB", "TB", "PB", "EB"]):
    """Returns a human readable string representation of NUM."""
    return (
        f"{num:.1f} {units[0]}"
        if num < 1024
        else human_size(num / 1000, units[1:])
    )


# with open('/data/sparc/sids5.list') as f:
#      specids = [int(line.strip()) for line in f if not line.startswith('#')]
def run_retrieve(specids, columns=None, xfer="p", verbose=True):
    #!print(f'Retrieving {len(specids):,} spectra')
    psutil.cpu_percent()  # begin interval
    client = SparclClient(url=rooturl)
    result = dict(numcols=len(columns), numspecids=len(specids))
    if verbose:
        print(f"Experiment: {pformat(result)}")
    tic()
    data = client.retrieve(specids, columns=columns, xfer=xfer)
    elapsed = toc()
    #!cpu = psutil.cpu_percent(interval=1)
    if verbose:
        print(f"len(specids)={len(specids)} len(data)={len(data)}")
    assert len(specids) == len(data)  # @@@ but some of ingest may have failed
    assert len(data[0]["spectra__coadd__flux"]) > 1000
    result.update(
        elapsed=elapsed,
        retrieved=len(data),
        rate=len(data) / elapsed,
        end_smrem=psutil.swap_memory().free,
        end_vmrem=psutil.virtual_memory().available,
        end_cpuload=os.getloadavg()[1],
        end_cpuperc=psutil.cpu_percent(),  # end interval
    )
    return result


def run_paged_retrieve(
    specids, columns=None, xfer="p", page=5000, verbose=True, keepall=False
):
    """Do 1 more more PAGE size retrieves to get data for all specids"""
    print(f"Paged Retrieve of {len(specids):,} spectra")
    psutil.cpu_percent()  # begin interval
    client = SparclClient(url=rooturl)
    result = dict(
        numcols=len(columns), numspecids=len(specids), xfer=xfer, page=page
    )
    if verbose:
        print(f"Experiment: {pformat(result)}")

    data = []
    datacnt = 0
    tic()
    for cnt in range(0, len(specids), page):
        pdata = client.retrieve(
            specids[cnt : cnt + page], columns=columns, xfer=xfer
        )
        datacnt += len(pdata)
        if keepall:
            data.extend(pdata)
    elapsed = toc()

    #! cpu = psutil.cpu_percent(interval=1)
    if verbose:
        print(f"len(specids)={len(specids)} datacnt={datacnt}")
    # assert len(specids) == len(data)   # @@@but some ingests may have failed
    #!assert len(data[0]['spectra__coadd__flux']) > 1000 # @@@
    result.update(
        elapsed=elapsed,
        retrieved=len(data),
        rate=datacnt / elapsed,
        end_smrem=psutil.swap_memory().free,
        end_vmrem=psutil.virtual_memory().available,
        end_cpuload=os.getloadavg()[1],
        end_cpuperc=psutil.cpu_percent(),  # end interval
    )
    return result


# flux,loglam,ivar,and_mask,or_mask,wdisp,sky,model
allcols = [
    "flux",
    "loglam",
    "ivar",
    "and_mask",
    "or_mask",
    "wdisp",
    "sky",
    "model",
]

experiment_0 = dict(
    xfers=["p"],
    specidcnts=[600, 60],
    numcols=range(1, 3),
)
experiment_1 = dict(
    xfers=["p"],
    specidcnts=[6, 60, 600, 6000, 30000],
    numcols=range(1, 3),
    # numcols=range(1,len(allcols)+1),
)
experiment_2 = dict(
    xfers=["p"],
    specidcnts=[1000, 100, 10],
    numcols=range(1, len(allcols) + 1),
)
experiment_3 = dict(
    xfers=["p"],
    specidcnts=[
        1000,
    ],
    numcols=reversed(range(1, len(allcols) + 1)),
)

experiment_8 = dict(
    xfers=[
        "p",
    ],
    specidcnts=[
        65000,
    ],
    numcols=[1, 2, 8],
)
experiment_9 = dict(
    xfers=["p", "j"],
    specidcnts=sorted(set([min(7 * 10**x, 65000) for x in range(6)])),
    numcols=range(1, len(allcols) + 1),
)


def run_trials(allspecids, verbose=True):
    # ex = experiment_9 #@@@
    ex = experiment_8  # @@@

    xfers = ex["xfers"]
    specidcnts = ex["specidcnts"]
    numcols = ex["numcols"]

    klist = ["elapsed", "numcols", "numspecids", "page", "rate", "xfer"]

    all = []
    for xfer in xfers:
        for n in numcols:
            cols = allcols[:n]
            for specidcnt in specidcnts:
                specids = allspecids[:specidcnt]
                #!result = run_retrieve(specids, columns=cols, xfer='p')
                result = run_paged_retrieve(specids, columns=cols, xfer="p")
                if verbose:
                    # print(f'Run-Result: {pformat(result)}')
                    reduced = dict(
                        (k, result[k]) for k in result.keys() if k in klist
                    )
                    print(f"Run-Result: {reduced}")
                all.append(result)
    report(all, len(allspecids), xfer=xfer)
    return all


def report(results, specidcnt, xfer=None, bandwidth=False):
    hostname, now = here_now()
    min1, min5, min15 = os.getloadavg()
    #!smrem = psutil.swap_memory().free
    #!vmrem = psutil.virtual_memory().available
    #!cpuperc = psutil.cpu_percent(interval=1)

    if bandwidth:
        pass
        #! s = speedtest.Speedtest()
        #! ul_speed = s.upload(threads=1)
        #! dl_speed = s.download(threads=1)
    else:
        #! ul_speed = 0
        dl_speed = 0

    #! Upload speed:    {human_size(ul_speed)}
    print(f"\nBenchmark run on {hostname} at {now} with {specidcnt} specids.")
    print(
        f"""
Transfer Method: {"Pickle" if xfer=='p' else "JSON"}
Download speed:  {human_size(dl_speed)}
    """
    )
    # Load Avg:        {min5:.1f}
    #         (avg num processes running over last 5 minutes)
    # CPU utilization: {cpuperc:.0f}%
    # Swap Mem Avail:    {human_size(smrem)}
    # Virtual Mem Avail: {human_size(vmrem)}
    # (Above statistics are for CLIENT.)

    #!print(f'Column\tSID\tRate \tAvg \tCPU \tSwap\tVirt')
    #!print(f' Count\tCnt\ts/sec\tLoad\tUtil\t Mem\t Mem')
    #!print(f'------\t---\t-----\t----\t----')
    print(f"Column\tSID\tRate ")
    print(f" Count\tCnt\ts/sec")
    print(f"------\t---\t-----")
    for r in results:
        print(
            (
                "{numcols}\t"
                "{numspecids}\t"
                "{rate:.0f}\t"
                #!"{end_cpuload:.02f}\t"
                #!"{end_cpuperc:.0f}%\t"
                #!"{smrem}\t"
                #!"{vmrem}\t"
            ).format(**r)
        )
        # smrem=human_size(r['end_smrem']),
        # vmrem=human_size(r['end_vmrem']),
    print(
        """
LEGEND:
  Rate:: spectra/second
  Transfer method:: Pickle, Json
  Load:: Number of processes in system run queue averaged over last 5 minutes.
  """
    )
    return "Done"


##############################################################################


def my_parser():
    parser = argparse.ArgumentParser(
        #!version='1.0.1',
        description="My shiny new python program",
        epilog='EXAMPLE: %(prog)s a b"',
    )
    allcols = [
        "flux",
        "loglam",
        "ivar",
        "and_mask",
        "or_mask",
        "wdisp",
        "sky",
        "model",
    ]
    #!dftcols = 'flux,loglam'
    dftcols = ",".join(allcols)
    parser.add_argument(
        "specids",
        type=argparse.FileType("r"),
        help=("File containing list of " "specobjids. One per line."),
    )
    parser.add_argument(
        "--cols",
        # choices=allcols,
        default=dftcols,
        help=(
            f"List of comma seperated columns to get. " f'Default="{dftcols}"'
        ),
    )
    parser.add_argument(
        "--xfer",
        default="p",
        help="Mode to use to transfer from Server to Client.",
    )

    parser.add_argument(
        "--loglevel",
        help="Kind of diagnostic output",
        choices=["CRTICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        default="WARNING",
    )
    return parser


def main():
    parser = my_parser()
    args = parser.parse_args()
    args.specids.close()
    args.specids = args.specids.name

    log_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(log_level, int):
        parser.error("Invalid log level: %s" % args.loglevel)
    logging.basicConfig(
        level=log_level,
        format="%(levelname)s %(message)s",
        datefmt="%m-%d %H:%M",
    )
    logging.debug("Debug output is enabled!!!")

    specids = []
    with open(args.specids, "r") as fin:
        for line in fin:
            if line.startswith("#"):
                continue
            specids.append(int(line.strip()))
    #! cols = args.cols.split(',')
    # print(f'specids count={len(specids)} columns={cols}')

    # run_retrieve(specids, columns=cols, xfer='p')
    print(f"Starting benchmark on {here_now()}")
    #! all = run_trials(specids)
    print(f"Finished benchmark on {here_now()}")


def foo(x):
    pass


if __name__ == "__main__":
    main()



"""Align or resample spectra related fields across multiple records."""

# See client.py for Doctest example
#
# For info about problems with floating point,
#   See:  https://docs.python.org/3/tutorial/floatingpoint.html
#   Also: https://docs.python.org/3/library/decimal.html#floating-point-notes
#
from decimal import Decimal

#
import numpy as np

#
import sparcl.client


# Map every wavelength of every record to index (ri,wi)
# where
#   ri: Record Index
#   wi: Window Index (offset of wavelength in WINDOW)
#   window: ordered list of wavelengths that include ALL unique
#           wavelengths in all records
#! def rec_woffset(records, window):
#!     ar = np.ones([len(records), len(window)])
#!     for ri, r in enumerate(records):
#!         for wl in r.wavelength:
#!             try:
#!                 wi = window.index(wl)
#!             except:
#!                 continue
#!             ar[ri,wi] = wl
#!     return ar


def _wavelength_offsets(records):
    # sorted list of wavelengths from ALL records
    window = sorted(
        set(records[0].wavelength).union(*[r.wavelength for r in records[1:]])
    )
    # offsets[ri] = index into WINDOW
    offsets = {
        ri: window.index(rec.wavelength[0]) for ri, rec in enumerate(records)
    }
    return (window, offsets)


def _validate_wavelength_alignment(records, window, offsets, precision=None):
    PLACES = Decimal(10) ** -precision if precision is not None else None
    #! print(f'DBG: PLACES={PLACES}')
    # Given an exact wavelength match between first wl (wavelength) in a rec
    # and the wl at its offset of WINDOW, ensure all the remaning wls
    # in rec match the next N wls of WINDOW.
    for ri, rec in enumerate(records):
        for wi, rwl in enumerate(rec.wavelength):  # wi=recwavelengthIndex
            if precision is None:
                recwl = Decimal(rwl)
            else:
                recwl = Decimal(rwl).quantize(PLACES)
            wwl = window[offsets[ri] + wi]
            #! msg = (f'Wavelength in '
            #!        f'Record[{ri}][{wi}] ({recwl}) does not match '
            #!        f'Window[{offsets[ri]+wi} = offset[{ri}]={offsets[ri]} '
            #!        f'+ {wi}]  ({wwl})'
            #!        )
            #! assert recwl == wwl, msg
            if recwl != wwl:
                msg = (
                    f"The spectra cannot be aligned with the given"
                    f' "precision" parameter ({precision}).'
                    f" Try lowering the precision value."
                )
                raise Exception(msg)


# We want to align a bunch of records by wavelength into a single
# 2d numpy array (record vs wavelength).  In general, we
# are not guaranteed that this is possible -- even if using only
# records from a single DataSet. So validate it first.
# (If not valid, allowing wavelength slop might help.)
def _align_wavelengths(records):
    window, offsets = _wavelength_offsets(records)
    _validate_wavelength_alignment(records, window, offsets)
    ar = np.ones([len(records), len(window)])
    for ri, r in enumerate(records):
        for wi, wl in enumerate(r.wavelength):
            ar[ri, offsets[ri + wi]] = wl  # @@@WRONG!!! We want FLUX
    return ar


def _tt1(numrecs=20, dr="BOSS-DR16"):
    client = sparcl.client.SparclClient()
    found = client.find(constraints=dict(data_release=[dr]), limit=numrecs)
    got = client.retrieve(found.ids)
    records = got.records
    window, offsets = _wavelength_offsets(records)
    print(f"Built window len={len(window)}; offsets={offsets}")
    # return records, window, offsets
    ar = _align_wavelengths(records)
    return ar


# precision:: number of decimal places
# "records" must contain "wavelength" field.
def _wavelength_grid_offsets(records, precision=11):
    PLACES = Decimal(10) ** -precision

    # set of wavelengths from ALL records. Quantized to precision
    gset = set()  # Grid SET
    for r in records:
        gset.update([Decimal(w).quantize(PLACES) for w in r.wavelength])
    grid = sorted(gset)  # 1D sorted list of wavelengths (bigger than any rec)
    #! print(f'DBG grid({len(grid)})[:10]={grid[:10]}')
    # offsets[ri] = index into GRID
    offsets = {
        ri: grid.index(Decimal(rec.wavelength[0]).quantize(PLACES))
        for ri, rec in enumerate(records)
    }
    return (grid, offsets)


# return 2D numpy array of FLUX values that is aligned to wavelength GRID.
# GRID is generally wider than flux for single record. Pad with NaN.
def _flux_grid(records, grid, offsets, precision=None):
    _validate_wavelength_alignment(records, grid, offsets, precision=precision)
    ar = np.full([len(records), len(grid)], np.nan)
    for ri, r in enumerate(records):
        for fi, flux in enumerate(r.flux):
            ar[ri, offsets[ri] + fi] = flux
    return ar


# RETURN 2D nparray(records,wavelengthGrid) = fieldValue
def _field_grid(records, fieldName, grid, offsets, precision=None):
    ar = np.full([len(records), len(grid)], np.nan)
    for ri, r in enumerate(records):
        for fi, fieldValue in enumerate(r[fieldName]):
            ar[ri, offsets[ri] + fi] = fieldValue
    return ar  # (wavelengthGrid, records)


# RETURN 2D nparray(fields,wavelengthGrid) = fieldValue
#! def rec_grid(rec, fields, grid, offsets, precision=None):
#!     ar = np.full([len(fields), len(grid)], np.nan)
#!     ri = 0
#!     for fi, fieldValue in enumerate(r[fieldName]):
#!         ar[ri, offsets[ri] + fi] = fieldValue
#!     return ar  # (wavelengthGrid, fields)


# Align flux from records into one array using quantization
#! def flux_records(records, precision=None):
#!     grid, offsets = wavelength_grid_offsets(records, precision=precision)
#!     ar = _flux_grid(records, grid, offsets, precision=precision)
#!     return ar, np.array([float(x) for x in grid])


def _validate_spectra_fields(records, fields):
    #! spectra_fields = [
    #!     client.fields.n2o["BOSS-DR16"][k]
    #!     for k, v in client.fields.attrs["BOSS-DR16"].items()
    #!     if v["storage"] == "S"
    #! ]
    [k for k in records[0].keys() if not k.startswith("_")]


# TOP level: Intended for access from Jupyter NOTEBOOK.
# Align spectra related field from records into one array using quantization.
def align_records(records, fields=["flux", "wavelength"], precision=7):
    """Align given spectra-type fields to a common wavelength grid.

    Args:
        records (list): List of dictionaries.
            The keys for all these dictionaries are Science Field Names.

        fields (:obj:`list`, optional): List of Science Field Names of
            spectra related fields to align and include in the results.
            DEFAULT=['flux', 'wavelength']

        precision (:obj:`int`, optional): Number of decimal points to use for
            quantizing wavelengths into a grid.
            DEFAULT=7

    Returns:
        tuple containing:
        - ar_dict(dict): Dictionary of 2D numpy arrays keyed by Field Name.
              Each array is shape: (numRecs, numzGridWavelengths)
        - grid(ndarray): 1D numpy array containing wavelength values.

    Example:
        >>> client = sparcl.client.SparclClient()
        >>> specflds = ['wavelength', 'model']
        >>> cons = {"data_release": ['BOSS-DR16']}
        >>> found = client.find(constraints=cons, limit=21)
        >>> got = client.retrieve(found.ids, include=specflds)
        >>> ar_dict, grid = align_records(got.records, fields=specflds)
        >>> ar_dict['model'].shape
        (21, 4669)

    """
    # Report Garbage In
    if "wavelength" not in fields:
        msg = (
            f'You must provide "wavelength" in the list provided'
            f' in the "fields" paramter.  Got: {fields}'
        )
        raise Exception(msg)
    if "wavelength" not in records[0]:
        msg = (
            f'Records must contain the "wavelength" field.'
            f" The first record contains fields: {sorted(records[0].keys())}"
        )
        raise Exception(msg)

    #! _validate_spectra_fields(records, fields)
    grid, offsets = _wavelength_grid_offsets(records, precision=precision)
    _validate_wavelength_alignment(records, grid, offsets, precision=precision)

    # One slice for each field; each slice a 2darray(wavelength, record)=fldVal
    adict = dict()
    for fld in fields:
        ar = _field_grid(records, fld, grid, offsets, precision=None)
        adict[fld] = ar

    return adict, np.array([float(x) for x in grid])


# with np.printoptions(threshold=np.inf, linewidth=210,
#   formatter=dict(float=lambda v: f'{v: > 7.3f}')): print(ar.T)  # noqa: E501

if __name__ == "__main__":
    import doctest

    doctest.testmod()



# End Users should not use anything from this file. All of it is
# considered experimentatal and be broken or change without notice.
############################################
# Python Standard Library
from urllib.parse import urlencode

#!from urllib.parse import urlparse
#!from warnings import warn
import pickle
import tempfile
import json

############################################
# External Packages
import requests

############################################
# Local Packages
#!from sparcl.fields import Fields
import sparcl.exceptions as ex

_STAGE = "https://sparclstage.datalab.noirlab.edu"  # noqa: E221
_PAT = "https://sparc1.datalab.noirlab.edu"  # noqa: E221

drs = ["SDSS-DR16", "BOSS-DR16", "DESI-EDR"]


def retrieve(
    ids,
    include=["id"],
    dataset_list=["BOSS-DR16"],
    server=_PAT,
    svc="spectras",  # or 'retrieve',
    limit=100,
    verbose=True,
):
    uparams = dict(
        include=",".join(include),
        limit=limit,
        dataset_list=",".join(dataset_list),
    )
    qstr = urlencode(uparams)

    url = f"{server}/sparc/{svc}/?{qstr}"
    if verbose:
        print(f"Using ids={ids[:2]}")
        print(f'Using url="{url}"')
        print(f"curl -X POST \"{url}\" -d '{json.dumps(ids)}' > retrieve.pkl")

    res = requests.post(url, json=ids)

    if res.status_code != 200:
        #! if verbose and ('traceback' in res.json()):
        #!     print(f'DBG: Server traceback=\n{res.json()["traceback"]}')
        raise ex.genSparclException(res, verbose=verbose)

    # unpack pickle file from result
    with tempfile.TemporaryFile(mode="w+b") as fp:
        for idx, chunk in enumerate(res.iter_content(chunk_size=None)):
            fp.write(chunk)
        # Position to start of file for pickle reading (load)
        fp.seek(0)
        results = pickle.load(fp)

    return results



# Python library
import datetime
import time
import socket
import itertools
import json
import subprocess

# External packages
#   none
# LOCAL packages
#   none


# data = {
#     "a": "aval",
#     "b": {
#         "b1": {
#             "b2b": "b2bval",
#             "b2a": {
#                 "b3a": "b3aval",
#                 "b3b": "b3bval"
#             }
#         }
#     }
# }
#
# data1 = AttrDict(data)
# print(data1.b.b1.b2a.b3b)  # -> b3bval
class _AttrDict(dict):
    """Dictionary subclass whose entries can be accessed by attributes
    (as well as normally).
    """

    def __init__(self, *args, **kwargs):
        def from_nested_dict(data):
            """Construct nested AttrDicts from nested dictionaries."""
            if not isinstance(data, dict):
                return data
            else:
                return _AttrDict(
                    {key: from_nested_dict(data[key]) for key in data}
                )

        super(_AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

        for key in self.keys():
            self[key] = from_nested_dict(self[key])


def tic():
    """Start tracking elapsed time. Works in conjunction with toc().

    Args:
       None.
    Returns:
       Elapsed time.
    """
    tic.start = time.perf_counter()


def toc():
    """Return elapsed time since previous tic().

    Args:
       None.
    Returns:
       Elapsed time since previous tic().
    """
    elapsed_seconds = time.perf_counter() - tic.start
    return elapsed_seconds  # fractional


def here_now():
    """Used to track info for benchmark. Probably OBE?

    Args:
       None.
    Returns:
       Time, date, and hostname.
    """
    hostname = socket.gethostname()
    now = str(datetime.datetime.now())
    return (hostname, now)


def objform(obj):
    """Nested structure of python object. Avoid spewing big lists.
    See also: https://code.activestate.com/recipes/577504/

    Args:
       obj: Python object.
    Returns:
       Length if list, objforms of dict contents if dict, type if
       anything else.
    Example:
       >>> res = client.sample_records(1)[0]
       >>> objform(res)
       <class 'sparcl.client.AttrDict'>
    """
    # dict((k,len(v)) for (k,v) in qs[0]['spzline'].items())
    if obj is None:
        return None
    elif type(obj) is list:
        if len(obj) > 999:
            return f"CNT={len(obj)}"
        elif len(obj) < 9:
            return [objform(x) for x in obj]
        else:
            return [objform(x) for x in obj[:10]] + ["..."]
    elif type(obj) is dict:
        return dict((k, objform(v)) for (k, v) in obj.items())
    else:
        return str(type(obj))


def dict2tree(obj, name=None, prefix=""):
    """Return abstracted nested tree. Terminals contain TYPE.
    As a special case, a list is given as a dict that represents a
    compound type.  E.G. {'<list(3835)[0]>': float} means a list of
    3835 elements where the first element is of type 'float'.  NB:
    Only the type of the first element in a list is given.  If the
    list has hetergeneous types, that fact is invisible in the
    structure!!
    """
    nextpfx = "" if name is None else (prefix + name + ".")
    showname = prefix if name is None else (prefix + name)
    if isinstance(obj, dict):
        children = dict()
        for k, v in obj.items():
            if isinstance(v, dict) or isinstance(v, list):
                val = dict2tree(v, name=k, prefix=nextpfx)
            else:
                #!val = {k: type(v).__name__}
                val = {nextpfx + k: type(v).__name__}
            children.update(val)
        tree = children if name is None else {showname: children}
    elif isinstance(obj, list):
        children = f"<list({len(obj)})[0]>:{type(obj[0]).__name__}"
        tree = {showname: children}
    else:
        tree = {showname: type(obj).__name__}
    return tree


def invLUT(lut):
    """Given dict[k]=v, Return dict[v]=k"""
    return {v: k for k, v in lut.items()}


def count_values(recs):
    """Count number of non-None values in a list of dictionaries.
    A key that exists with a value of None is treated the same as a
    key that does not exist at all. i.e. It does not add to the count.

    Args:
       recs (:obj:`list`): ('records') List of dictionaries.

    Returns:
        A dictionary. Keys are the full list of keys available in any
        of the recs.  Values are the count of occurances of non-None values
        for that key.

    >>> count_values([dict(a=None, b=3), dict(a=1, b=2), dict(a=None, b=2)])
    {'a': 1, 'b': 3}
    """
    allkeys = set(list(itertools.chain(*recs)))
    return {k: sum(x.get(k) is not None for x in recs) for k in allkeys}


# In case I want to give CURL equivalents for client methods.
#
# Retrieve may return results as a pickle file since it usually contains
# spectra vectors.  To handle pickle results, write curl output to out.pkl
# and do something like:
#   with open('out.pkl', 'rb') as f: res = pickle.load(f)
def curl_retrieve_str(ids, server, svc="spectras", qstr=None):
    #! ids = ['00000dd7-b1ff-48ed-b162-46d9d65f829c', 'BADID']
    #!svc = 'spectras' if use_async else 'retrieve'
    #! qstr = urlencode(uparams)
    # server = "https://sparc1.datalab.noirlab.edu"
    qqstr = "" if qstr is None else f"?{qstr}"
    url = f"{server}/sparc/{svc}/{qqstr}"
    curlpost1 = "curl -X 'POST' -H 'Content-Type: application/json' "
    curlpost2 = f"-d '{json.dumps(ids)}' '{url}'"
    curlpost3 = " | python3 -m json.tool"
    return curlpost1 + curlpost2 + curlpost3


# see: curl_retrieve_str
def curl_find_str(sspec, server, qstr=None):
    qqstr = "" if qstr is None else f"?{qstr}"
    url = f"{server}/sparc/find/{qqstr}"
    curlpost1 = "curl -X 'POST' -H 'Content-Type: application/json' "
    curlpost2 = f"-d '{json.dumps(sspec)}' '{url}'"
    curlpost3 = " | python3 -m json.tool"
    return curlpost1 + curlpost2 + curlpost3


def githash(verbose=False):
    try:
        #  "/usr/bin/git"
        ret = subprocess.run(["git", "rev-parse", "HEAD"], capture_output=True)
        commit_hash = ret.stdout.decode().strip()
    except Exception as err:
        if verbose:
            print(err)
        commit_hash = "<NA>"
    return commit_hash
